@@ -1,16 +1,35 @@
 //! Types and traits for converting objects to addresses which
 //! Tor can connect to.
 
+use crate::err::ErrorDetail;
 use std::net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr, SocketAddrV4, SocketAddrV6};
 use std::str::FromStr;
 use thiserror::Error;
 
 // ----------------------------------------------------------------------
 
-/// An object that can be converted to a [`TorAddr`] with a minimum
-/// of risk.
+/// An object that can be converted to a [`TorAddr`] with a minimum of risk.
+///
+/// Typically, this trait will be implemented for a hostname or service name.
+///
+/// Don't implement this trait for IP addresses and similar types; instead,
+/// implement [`DangerouslyIntoTorAddr`] for those.  (The trouble with accepting
+/// IP addresses is that, in order to get an IP address, most programs will do a
+/// local hostname lookup, which will leak the target address to the DNS
+/// resolver. The `DangerouslyIntoTorAddr` trait provides a contract for careful
+/// programs to say, "I have gotten this IP address from somewhere safe."  This
+/// trait is for name-based addressing and similar, which _usually_ gets its
+/// addresses from a safer source.)
 ///
 /// [*See also: the `TorAddr` documentation.*](TorAddr)
+///
+/// # Design note
+///
+/// We use a separate trait here, instead of using `Into<TorAddr>` or
+/// `TryInto<TorAddr>`, because `IntoTorAddr` implies additional guarantees
+/// relating to privacy risk.  The separate trait alerts users that something
+/// tricky is going on here, and encourages them to think twice before
+/// implementing `IntoTorAddr` for their own types.
 pub trait IntoTorAddr {
     /// Try to make a [`TorAddr`] to represent connecting to this
     /// address.
@@ -29,8 +48,11 @@ pub trait IntoTorAddr {
 ///
 /// [*See also: the `TorAddr` documentation.*](TorAddr)
 pub trait DangerouslyIntoTorAddr {
-    /// Try to make a [`TorAddr`] to represent connecting to this
-    /// address.
+    /// Try to make a [`TorAddr`] to represent connecting to `self`.
+    ///
+    /// By calling this function, the caller asserts that `self` was
+    /// obtained from some secure, private mechanism, and **not** from a local
+    /// DNS lookup or something similar.
     fn into_tor_addr_dangerously(self) -> Result<TorAddr, TorAddrError>;
 }
 
@@ -99,7 +121,7 @@ pub struct TorAddr {
 }
 
 impl TorAddr {
-    /// Construct a TorAddr from its consituent parts, rejecting it if the
+    /// Construct a TorAddr from its constituent parts, rejecting it if the
     /// port is zero.
     fn new(host: Host, port: u16) -> Result<Self, TorAddrError> {
         if port == 0 {
@@ -147,17 +169,17 @@ impl TorAddr {
     pub(crate) fn enforce_config(
         &self,
         cfg: &crate::config::ClientAddrConfig,
-    ) -> Result<(), crate::Error> {
+    ) -> Result<(), ErrorDetail> {
         if !cfg.allow_local_addrs && self.is_local() {
-            return Err(crate::Error::LocalAddress);
+            return Err(ErrorDetail::LocalAddress);
         }
 
         if let Host::Hostname(addr) = &self.host {
             if !is_valid_hostname(addr) {
-                return Err(crate::Error::InvalidHostname);
+                return Err(ErrorDetail::InvalidHostname);
             }
             if addr.to_lowercase().ends_with(".onion") {
-                return Err(crate::Error::OnionAddressNotSupported);
+                return Err(ErrorDetail::OnionAddressNotSupported);
             }
         }
 
@@ -383,8 +405,8 @@ mod test {
 
     #[test]
     fn validate_addr() {
-        use crate::Error;
-        fn val<A: IntoTorAddr>(addr: A) -> Result<TorAddr, Error> {
+        use crate::err::ErrorDetail;
+        fn val<A: IntoTorAddr>(addr: A) -> Result<TorAddr, ErrorDetail> {
             let toraddr = addr.into_tor_addr()?;
             toraddr.enforce_config(&Default::default())?;
             Ok(toraddr)
@@ -399,17 +421,20 @@ mod test {
 
         assert!(matches!(
             val("-foobar.net:443"),
-            Err(Error::InvalidHostname)
+            Err(ErrorDetail::InvalidHostname)
         ));
         assert!(matches!(
             val("www.torproject.org"),
-            Err(Error::Address(TorAddrError::NoPort))
+            Err(ErrorDetail::Address(TorAddrError::NoPort))
         ));
 
-        assert!(matches!(val("192.168.0.1:80"), Err(Error::LocalAddress)));
+        assert!(matches!(
+            val("192.168.0.1:80"),
+            Err(ErrorDetail::LocalAddress)
+        ));
         assert!(matches!(
             val("eweiibe6tdjsdprb4px6rqrzzcsi22m4koia44kc5pcjr7nec2rlxyad.onion:443"),
-            Err(Error::OnionAddressNotSupported)
+            Err(ErrorDetail::OnionAddressNotSupported)
         ));
     }
 
